/* pal.c -- 4bpp/8bpp paletted texture example -- Jim Ursetto (jim@3e8.org) */

#include <kos.h>
#include <stdlib.h>

#define PAL_FMT_ARGB1555 0
#define PAL_FMT_RGB565 1
#define PAL_FMT_ARGB4444 2
#define PAL_FMT_ARGB8888 3

/* We have to diddle with the poly header ourselves after
 * creation, since KOS doesn't support palette bank select. */
#define PAL_BANK_SHIFT_8BPP (25)
#define PAL_BANK_SHIFT_4BPP (21)
#define PAL_BANK_MASK_8BPP  (3 << PAL_BANK_SHIFT_8BPP)
#define PAL_BANK_MASK_4BPP (63 << PAL_BANK_SHIFT_4BPP)

/* If undefined, use 4bpp palette (which is more interesting anyway). */
#undef USE_PAL8BPP

/* 4 banks of 256 entries in 8bpp palette, 64 banks of 16 entries in 4bpp */
#ifdef USE_PAL8BPP
#define MAX_SPRITES 4
unsigned int (*palette)[4][256] = (unsigned int (*)[4][256])0xa05f9000;
#else
#define MAX_SPRITES 64
unsigned int (*palette)[64][16] = (unsigned int (*)[64][16])0xa05f9000;
#endif

#define WIDTH  640
#define HEIGHT 480

typedef struct {
	float x, y, z;
	float dx, dy;
	float s, ds, smax;
	uint32 c;
} sprite_t;

void draw_textured_pal(float x, float y, float z, int bank, float scale);

int randnum(int limit) { return (rand() & 0x7fffffff) % limit; }

uint32 txr_face_twid;
uint8 txr_face[256] = 
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 5, 5, 1, 1, 5, 5, 1, 0, 0, 0, 0,
  0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
  0, 0, 1, 1, 1, 2, 1, 2, 2, 1, 2, 1, 1, 1, 0, 0,
  0, 0, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 0, 0,
  0, 0, 1, 1, 1, 2, 1, 2, 2, 1, 2, 1, 1, 1, 0, 0,
  0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0,
  0, 0, 1, 1, 1, 3, 3, 3, 3, 3, 3, 1, 1, 1, 0, 0,
  0, 0, 1, 1, 1, 1, 1, 1, 4, 4, 4, 1, 1, 1, 0, 0,
  0, 0, 0, 1, 1, 1, 1, 1, 4, 4, 4, 1, 1, 0, 0, 0,
  0, 0, 0, 0, 1, 1, 1, 1, 1, 4, 1, 1, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

uint32 txr_face_twid_4bpp;
uint8 txr_face_4bpp[128];

int main(void)
{
	uint8 maple_controller;
	cont_cond_t cond;
	int i;
	int c, dc, p;
	sprite_t faces[MAX_SPRITES];

	kos_init_all(IRQ_ENABLE | TA_ENABLE, ROMDISK_NONE);

	maple_controller = maple_first_controller();

	/* Convert texture from 8 to 4bpp. */
	/* Reverse the order of the txr_face[] operands
	 * for an amusing result. */
	for (i = 0; i < 16 * 16 / 2; i++) {
//		txr_face_4bpp[i] = (txr_face[i*2] << 4) | (txr_face[i*2+1]);
		txr_face_4bpp[i] = (txr_face[i*2+1] << 4) | (txr_face[i*2]);
	}

	txr_face_twid = ta_txr_allocate(16 * 16);
	txr_twiddle_copy_general(txr_face, txr_face_twid, 16, 16, 8);
	txr_face_twid_4bpp = ta_txr_allocate(16 * 16 / 2);
	txr_twiddle_copy_general(txr_face_4bpp, txr_face_twid_4bpp, 16, 16, 4);

	/* This is not very random (same seed every time). */
	srand(123);
	for (i = 0; i < MAX_SPRITES; i++) {
		faces[i].x = (float)randnum(WIDTH);
		faces[i].y = (float)randnum(HEIGHT);
		faces[i].z = (float)(randnum(512) + 1);
		faces[i].dx = (float)(randnum(30) - (30 / 2)) / 10.0f; // -1.5f to +1.5f
		faces[i].dy = (float)(randnum(30) - (30 / 2)) / 10.0f;
		faces[i].s = (float)(randnum(64) + 10);
		faces[i].c = (float)(randnum(4096));
		faces[i].smax = faces[i].s * randnum(30) / 10.0f;
		faces[i].ds = (float)(randnum(30) - (30 / 2)) / 10.0f;
	}

	// Set the palette entry pixel format through the Palette RAM CTRL register.
	*(volatile uint32 *)0xa05f8108 = PAL_FMT_ARGB4444;

	for (i = 0; i < MAX_SPRITES; i++) {
		(*palette)[i][0] = 0x0000;  //transparent
		(*palette)[i][1] = 0x7fef;  //yellow
		(*palette)[i][2] = 0x0000;
		(*palette)[i][3] = 0x0000;
		(*palette)[i][4] = 0x7c00;  //red
		(*palette)[i][5] = 0x0000;
	}

	p = 0;
	c = 0; dc = 2;
	while (1) {
		if (cont_get_cond(maple_controller, &cond) < 0)
			break;
		if (!(cond.buttons & CONT_START))
			break;
		pvr_dummy_poly(TA_OPAQUE);
		ta_commit_eol();
		pvr_dummy_poly(TA_TRANSLUCENT);
		for (i = 0; i < MAX_SPRITES; i++) {
			draw_textured_pal(faces[i].x, faces[i].y, faces[i].z, i, faces[i].s);
		}
		ta_commit_eol();
		ta_finish_frame();
		for (i = 0; i < MAX_SPRITES; i++) {
			faces[i].x += faces[i].dx;
			faces[i].y += faces[i].dy;
			faces[i].s += faces[i].ds;
			if (faces[i].x >= WIDTH || faces[i].x <= 0) { faces[i].dx = -faces[i].dx; faces[i].x += faces[i].dx; }
			if (faces[i].y >= HEIGHT || faces[i].y <= 0) { faces[i].dy = -faces[i].dy; faces[i].y += faces[i].dy; }
			if (faces[i].s >= faces[i].smax || faces[i].s <= 1) { faces[i].ds = -faces[i].ds; faces[i].s += faces[i].ds; }
			if (p < 600 * 60) {	
				// vary color shift between 0 and 7 based on time; sync fadeout with color shift
				(*palette)[i][1] = ((c/16) << 12) | (i << ((p >> 8) & 7)); 
			} else {
				(*palette)[i][1] = ((c/8) << 10) | 0x3ef; // Not as smooth a fade as I want
			}  
		}
		p++;
		c+=dc;
		if (c >= 255 || c < 0) { dc = -dc; c+= dc; }
	}
	kos_shutdown_all();
	return 0;
}

void draw_textured_pal(float x, float y, float z, int bank, float scale)
{
	poly_hdr_t poly_textured;
	vertex_ot_t vert;
	float x1 = x - scale / 2.0f, y1 = y - scale / 2.0f, x2 = x1 + scale, y2 = y1 + scale;
	float u1 = 0.0f, v1 = 0.0f, u2 = 1.0f, v2 = 1.0f;

#ifdef USE_PAL8BPP
	ta_poly_hdr_txr(&poly_textured, TA_TRANSLUCENT, TA_PAL8BPP, 16, 16, txr_face_twid, TA_BILINEAR_FILTER);
	// Bank select -- KOS doesn't let us do this directly.
	poly_textured.flags4 &= ~PAL_BANK_MASK_8BPP;
	poly_textured.flags4 |= bank << PAL_BANK_SHIFT_8BPP;
#else
	ta_poly_hdr_txr(&poly_textured, TA_TRANSLUCENT, TA_PAL4BPP, 16, 16, txr_face_twid_4bpp, TA_BILINEAR_FILTER);
	poly_textured.flags4 &= ~PAL_BANK_MASK_4BPP;
	poly_textured.flags4 |= bank << PAL_BANK_SHIFT_4BPP;
#endif
	ta_commit_poly_hdr(&poly_textured);

	vert.flags = TA_VERTEX_NORMAL;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	vert.oa = 1.0f; vert.or = 1.0f; vert.og = 1.0f; vert.ob = 1.0f;
	vert.z = 512.0f;
	vert.x = x1; vert.y = y2;
	vert.u = u1; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x1; vert.y = y1;
	vert.u = u1; vert.v = v1;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y2;
	vert.u = u2; vert.v = v2;
	ta_commit_vertex(&vert, sizeof(vert));
	vert.x = x2; vert.y = y1;
	vert.u = u2; vert.v = v1;
	vert.flags = TA_VERTEX_EOL;
	ta_commit_vertex(&vert, sizeof(vert));
}
